﻿package  {
	
	import com.yoambulante.midifiles.MidiFilesReader;
	import com.yoambulante.midifiles.MidiMessage;
	import com.yoambulante.midifiles.MidiTrack;	
	import com.yoambulante.midifiles.SamplerKey;
	import com.yoambulante.midifiles.SamplesExtractor;
	import flash.display.MovieClip;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.events.ProgressEvent;
	import flash.events.SampleDataEvent;
	import flash.geom.Point;
	import flash.media.Sound;
	import flash.media.SoundChannel;
	import flash.utils.ByteArray;
	import flash.utils.setTimeout;
	
	
	public class Pagode extends MovieClip {
		public var btn:MovieClip;		
		private var playing:Boolean;
		private var sound:Sound;
		private var soundChannel:SoundChannel;
		private var reader:MidiFilesReader;
		
		[Embed(source = 'pagodinho.mid', mimeType = 'application/octet-stream')]
		private var MidiFile:Class;
		
		private var extractors:Vector.<SamplesExtractor>; //these objects convert an flash.media.Sound into stereo samples in form of Points
		private var extractIndex:uint;		
		
		private var instruments:Vector.<Instrument>; //all 6 instruments
		private var transpose:Vector.<int>; //transpose value for all 6 instruments
		
		private var spqn:Number; //seconds per quater note
		private var tpqn:uint; //ticks per quarter note
		private var mspqn:uint; //microseconds per quater note		
		private var totalTracks:uint; //total amount of tracks
		private var totalTicks:uint; //how many ticks the whole song has, in future it could be used for jumping time
		private var acc_time:Vector.<int>; //accumulative time between MIDI events.
		private var vols:Vector.<Slider>;
		
		public function Pagode() {
			// constructor code	
			totalTracks = 0;
			
			btn.gotoAndStop(1);
			playing = false; //it help to know when audio is playing.
			
			sound = new Sound();
			sound.addEventListener(SampleDataEvent.SAMPLE_DATA, generateSound);
			
			reader = new MidiFilesReader();			
			reader.addEventListener(MidiFilesReader.ON_DATA_ERROR, midiError);
			reader.addEventListener(MidiFilesReader.ON_HEADER_LOADED, midiValidHeader);
			reader.addEventListener(MidiFilesReader.ON_DATA_LOADED, midiReady);			
			
			vol0.value = .7; //pandeiro de couro volume
			vol3.value = .4; //can of beer volume
			vol4.value = .7; //cabasa volume
			tempoSlider.label.text = "?";
			tempoSlider.addEventListener(Event.CHANGE, tempoChange);
			
			var i:int;
			transpose = new Vector.<int>(6, true);
			for (i = 0; i < 6; i++) {
				transpose[i] = 0;
				this["trans" + i].addEventListener(Event.CHANGE, transposeChange);
			}
			
			instruments = new Vector.<Instrument>(6, true);
			instruments[0] = new Instrument(0, 3); //Pandeiro Couro channel 1, 3 samples
			instruments[1] = new Instrument(1, 18); //Pandeiro Nylon channel 2, 18 samples
			instruments[2] = new Instrument(2, 3); //Surdo channel 3, 3 samples
			instruments[3] = new Instrument(3, 9); //Can of beer channel 4, 9 samples			
			instruments[4] = new Instrument(4, 1); //Cabasa channel 5, only 1 sample						
			instruments[5] = new Instrument(5, 59); //Cavaquinho channel 6, 59 samples
			//create sound extractors			
			extractors = new Vector.<SamplesExtractor>(93, true);			
			extractors[0] = new SamplesExtractor(getSound("cabasa"));						
			extractors[1] = new SamplesExtractor(getSound("surdo01"));
			extractors[2] = new SamplesExtractor(getSound("surdo02"));
			extractors[3] = new SamplesExtractor(getSound("surdo03"));			
			extractors[4] = new SamplesExtractor(getSound("PandeiroCouro1"));
			extractors[5] = new SamplesExtractor(getSound("PandeiroCouro2"));
			extractors[6] = new SamplesExtractor(getSound("PandeiroCouro3"));			
			for (i = 1; i <= 9; i++) {
				extractors[6+i] = new SamplesExtractor(getSound("CanOfBeer0"+i));	
			}
			for (i = 1; i <= 18; i++) {
				extractors[15+i] = new SamplesExtractor(getSound("PandeiroNylon"+(i<10?"0":"")+i));	
			}
			for (i = 1; i <= 59; i++) {
				extractors[33+i] = new SamplesExtractor(getSound("cavaco"+(i<10?"0":"")+i));	
			}
			
			//volume controls are only used when we press the keys per channel, it affects velocity of the key.
			vols = new Vector.<Slider>(7, true);
			for (i = 0; i < 6; i++) {
				vols[i] = this["vol" + i];
			}
			vols[6] = master; //master control is always used when pressing keys.
			
			extractIndex = 0;
			for (i = 0; i < extractors.length; i++) {
				extractors[i].addEventListener(ProgressEvent.PROGRESS, extractingSamples);
				extractors[i].addEventListener(Event.COMPLETE, samplesReady);	
			}
			
			btn.addEventListener(MouseEvent.CLICK, extractSamples);
			btn.buttonMode = true;
			
		}
		
		private function transposeChange(e:Event):void {
			var m:SliderEx = e.currentTarget as SliderEx;
			var i:int = int(String(m.name).substr(5));
			transpose[i] = Math.round((m.value-.5)/.5 * 3);
			m.label.text = (transpose[i] >= 0 ? "+" : "") + transpose[i]; 
		}
		
		
		private function getSound(linkage:String):Sound {			
			var C:Class = this.loaderInfo.applicationDomain.getDefinition(linkage+".wav") as Class;
			if (C) {			
				return new C as Sound;
			}
			return null; //error sound in library not found
		}
		
		private function extractSamples(e:MouseEvent):void {
			btn.removeEventListener(MouseEvent.CLICK, extractSamples);
			btn.buttonMode = false;
			extractors[0].extract(); //start extracting.
		}
		
		private function getDecimal(n:Number):String { //get two decimals in string format of a rounded Number.
			var s:String = String(Math.round(n * 10000) / 10000 );
			var a:Array = s.split(".");
			if (a.length == 1) {
				a.push("00");
			} else if (a[1].length == 1) {
				a[1] = a[1] + "0";
			}
			return a[0]+"."+String(a[1]).substr(0,2);
		}
		
		//------------------- EXTRACTOR EVENTS [BEGIN]--------------------//
		private function extractingSamples(e:ProgressEvent):void {
			//show progress when the audio is big and can take a while.
			status.text = "extracting samples... "+getDecimal((extractIndex+1)/extractors.length*100) +"%";
		}
		
		private function samplesReady(e:Event):void {
			//create the keyboard notes.
			if (extractIndex == 0) { //is Cabasa?				
				instruments[4].setSamplerKey(0, 
												new MySamplerKey(extractors[extractIndex], -3, 60, 4),
												new MySamplerKey(extractors[extractIndex], -2, 60, 4),
												new MySamplerKey(extractors[extractIndex], -1, 60, 4),
												new MySamplerKey(extractors[extractIndex], 0, 60, 4),
												new MySamplerKey(extractors[extractIndex], 1, 60, 4),
												new MySamplerKey(extractors[extractIndex], 2, 60, 4),
												new MySamplerKey(extractors[extractIndex], 3, 60, 4)); //(7 + 24) - i, 88 - i used in previous example
			} else if (extractIndex >= 1 && extractIndex <= 3) { //is Surdo?
				instruments[2].setSamplerKey(extractIndex - 1, 
												new MySamplerKey(extractors[extractIndex], -3, 60, 2),
												new MySamplerKey(extractors[extractIndex], -2, 60, 2),
												new MySamplerKey(extractors[extractIndex], -1, 60, 2),
												new MySamplerKey(extractors[extractIndex], 0, 60, 2),
												new MySamplerKey(extractors[extractIndex], 1, 60, 2),
												new MySamplerKey(extractors[extractIndex], 2, 60, 2),
												new MySamplerKey(extractors[extractIndex], 3, 60, 2));
			} else if (extractIndex >= 4 && extractIndex <= 6) { //Pandeiro Couro?
				instruments[0].setSamplerKey(extractIndex - 4, 
												new MySamplerKey(extractors[extractIndex], -3, 60, 0),
												new MySamplerKey(extractors[extractIndex], -2, 60, 0),
												new MySamplerKey(extractors[extractIndex], -1, 60, 0),
												new MySamplerKey(extractors[extractIndex], 0, 60, 0),
												new MySamplerKey(extractors[extractIndex], 1, 60, 0),
												new MySamplerKey(extractors[extractIndex], 2, 60, 0),
												new MySamplerKey(extractors[extractIndex], 3, 60, 0));
			} else if (extractIndex >= 7 && extractIndex <= 15) { //Can of Beer?
				instruments[3].setSamplerKey(extractIndex - 7, 
												new MySamplerKey(extractors[extractIndex], -3, 60, 3),
												new MySamplerKey(extractors[extractIndex], -2, 60, 3),
												new MySamplerKey(extractors[extractIndex], -1, 60, 3),
												new MySamplerKey(extractors[extractIndex], 0, 60, 3),
												new MySamplerKey(extractors[extractIndex], 1, 60, 3),
												new MySamplerKey(extractors[extractIndex], 2, 60, 3),
												new MySamplerKey(extractors[extractIndex], 3, 60, 3));
			} else if (extractIndex >= 16 && extractIndex <= 33) { //Pandeiro Nylon?
				instruments[1].setSamplerKey(extractIndex - 16, 
												new MySamplerKey(extractors[extractIndex], -3, 60, 1),
												new MySamplerKey(extractors[extractIndex], -2, 60, 1),
												new MySamplerKey(extractors[extractIndex], -1, 60, 1),
												new MySamplerKey(extractors[extractIndex], 0, 60, 1),
												new MySamplerKey(extractors[extractIndex], 1, 60, 1),
												new MySamplerKey(extractors[extractIndex], 2, 60, 1),
												new MySamplerKey(extractors[extractIndex], 3, 60, 1));
			} else if (extractIndex >= 34) { //Cavaquinho?
				instruments[5].setSamplerKey(extractIndex - 34, 
												new MySamplerKey(extractors[extractIndex], -3, 60, 5),
												new MySamplerKey(extractors[extractIndex], -2, 60, 5),
												new MySamplerKey(extractors[extractIndex], -1, 60, 5),
												new MySamplerKey(extractors[extractIndex], 0, 60, 5),
												new MySamplerKey(extractors[extractIndex], 1, 60, 5),
												new MySamplerKey(extractors[extractIndex], 2, 60, 5),
												new MySamplerKey(extractors[extractIndex], 3, 60, 5));
			}
			if (++extractIndex < extractors.length) {
				extractors[extractIndex].extract();
			} else {
				//read the midi file content.
				reader.loadFromBytes(new MidiFile as ByteArray); //note you can use both reader.loadFromBytes or reader.loadFile for loading external files.
			}
		}
		
		private function click(e:MouseEvent):void {
			//click on the button			
			if (!playing){
				playing = true;
				btn.gotoAndStop(2);
				soundChannel = sound.play();
			} else {
				//clean up previous note if there is any
				while (playingNotes.length > 0) {
					playingNotes.splice(0, 1);
				}
				totalPlayingNotes = 0;
				btn.gotoAndStop(1);
				soundChannel.stop();
				playing = false;
				btn.removeEventListener(MouseEvent.CLICK, click); //enabled button.
				btn.buttonMode = false;
				//reload midi file in case the user wnat to played it again.
				reader.loadFromBytes(new MidiFile as ByteArray); //note you can use both reader.loadFromBytes or reader.loadFile for loading external files.
				//sound.removeEventListener(SampleDataEvent.SAMPLE_DATA, generateSound);
				//sound = new Sound(); //create a new sound. Flash Player always keeps some audio data it is previous buffer when stop.
				//sound.addEventListener(SampleDataEvent.SAMPLE_DATA, generateSound);
				soundChannel = null;
			}
		}
		//------------------- EXTRACTOR EVENTS [END]--------------------//
		
		//------------------- MIDI READER EVENTS [BEGIN] --------------------//
		private function midiError(e:Event):void {
			//there was an error loading midi file
			trace("wtf?");
		}
		
		private function midiValidHeader(e:Event):void {
			//when midi header was loaded load all events.
			while (reader.nextEvent()); //depending on how big your midi file is you may need to make your own wait dialog
		}
		
		private function midiReady(e:Event):void {			
			spqn = .5;			
			//extract time signature
			totalTracks = reader.totalTracks;
			
			var lastAccDTime:uint = 0;
			var accDTime:uint;			
			var trk:MidiTrack;
			var midi_event:MidiMessage;
			trace("reading tracks... " + reader.getTrackTypeDescription());
			for (var t:int = 0; t < totalTracks; t++) {				
				trk = reader.getTrack(t);					
				if (trk) {					
					midi_event = null;					
					var setTime:MidiMessage = null;												
					accDTime = 0;
					for (i = 0; i < trk.totalEvents; i++) { //go trhough the events.					
						midi_event = trk.getEvent(i); //use midi_event.toString() to know more about the event
						accDTime += midi_event.delta_time;						
						if (accDTime > lastAccDTime) {
							lastAccDTime = accDTime;
						}
						if (midi_event.category == MidiMessage.CATEGORY_METAEVENT) {
							if (midi_event.type == 81) { //Set Tempo								
								setTime = midi_event;								
							}
						}						
					}
					if (setTime) {
						tpqn = reader.resolution; //sequencer resolution, how many ticks contains a quarter note.											
						mspqn = setTime.param1;						
						spqn = mspqn / 1000000;	
						tempoSlider.label.text = Math.round(60 / spqn).toString();
						tempoSlider.value = (Math.round(60 / spqn) - 85) / 50;
						trace("SET TIME, ticks per quarter note:" + tpqn + ", microseconds per quarter note:" + mspqn + ", seconds per quarter note:" + spqn);																								
						acc_time = new Vector.<int>(); //reset the accumulated time.						
						for (i = 0; i < totalTracks; i++) {
							acc_time[i] = -1;
						}
						//we got the time, so we can start playing the midi file!						
						btn.addEventListener(MouseEvent.CLICK, click); //enabled button.
						btn.buttonMode = true;
						//break; 
					}
				}
			}
			//remove all events before the first note in all tracks.
			//trace("READY! >>> total ticks : " + lastAccDTime + ", time length: " + Math.floor((lastAccDTime / tpqn) * spqn) + " secs.");			
			totalTicks = lastAccDTime;
			status.text =  "READY TO PLAY, total ticks : " + lastAccDTime + ", time length: " + getDecimal((totalTicks / tpqn * spqn)/60) + " mins"
		}
		//------------------- MIDI READER EVENTS [END] --------------------//
		
		
		//------------------- DYNAMIC AUDIO GENERATION [BEGIN]--------------------//
		private var totalPlayingNotes:uint = 0; //how many notes simultaneously are we playing
		private var playingNotes:Vector.<MySamplerKey> = new Vector.<MySamplerKey>(); //current notes simultaneously playing
		/*private function getNodeByMidiNote(channel:uint, midiNote:uint):MySamplerKey {
			for (var i:int = 0; i < totalNotes; i++) {
				if (notes[channel][i].midiNote == midiNote) {
					return notes[channel][i];
				}
			}
			return null; //error, no key found
		}*/
		
		private function isPlayingNote(note:MySamplerKey):Boolean {			 
			for (var k:uint = 0; k < totalPlayingNotes; k++) { //k is only used 
				if (note == playingNotes[k]) {
					return true;
				}
			}
			return false;
		}
		
		private function tempoChange(e:Event):void {
			setBPM(tempoSlider.value);
		}
		
		private function setBPM(dec:Number):void {			
			//A = 60 / (X / 1000000). example: 120 = 60 / (500000 / 1000000)			
			//X = (60 / A) * 1000000
			var temp:Number = spqn; //catch before
			var bpm:Number = dec * 50 + 85; //between 85 and 135			
			spqn = 60 / bpm;
			tempoSlider.label.text = Math.round(60 / spqn).toString();
			status.text =  "new time length: " + getDecimal((totalTicks / tpqn * spqn) / 60) + " mins";
			for (i = 1; i < totalTracks; i++) {
				acc_time[i] *= spqn / temp;
			}
			for (i = 1; i < totalTracks; i++) {
				trk = reader.getTrack(i);				
				while (trk.totalEvents > 0){
					event = trk.getEvent(0); //affect next event delta time
					secs = acc_time[i] * STS;				
					var event_sec:Number = (event.delta_time / tpqn) * spqn;					
					if (secs >= event_sec) {
						//event must be remove
						acc_time[i] = int((secs - event_sec) / STS);						
						trk.removeEvent(0);						
					} else {
						break;
					}
				}				
			}
			
		}
		
		private const TOTAL_SAMPLES:uint = 2048; //how many samples we want to provide to the sound buffer?
		private const STS:Number = 1 / 44100; //How many seconds is a sample? helps to ranslate 1 sample to seconds.
		private const sample:Point = new Point(); //use to write the output.
		private var i:uint; //looping through audio samples (44100 per seconds)
		private var t:uint; //total amount of events remaining.
		private var n:uint; //used when filling the sample buffer with playingNotes
		private var trk:MidiTrack;
		private var secs:Number; //current 			
		private var maxVel:Vector.<Number> = new Vector.<Number>(6,true); //used for showing the volume level.
		private var maxVelAll:Number; //Master volume level.
		private var vol:Number;
		private var event:MidiMessage;		
		private var mustPlayNote:Boolean; //used for detecting when we have to play the note or not, it can't be too early, not too soon.		
		private var note:MySamplerKey;
		private var note_sample:Point; //temporary sample value.				
		
		
		private function generateSound(e:SampleDataEvent):void {						
			for (i = 0; i < 6; i++) {
				maxVel[i] = 0; //reset volumes
			}
			for (i = 0; i < TOTAL_SAMPLES; i++) { //samples to generate				
				for (n = 1; n < totalTracks; n++) {			
					trk = reader.getTrack(n);
					acc_time[n]++; //accumulated samples per second since last event.
					secs = acc_time[n] * STS;
					t = trk.totalEvents; //total events.										
					while (t > 0){ //read events.		
						event = trk.getEvent(0);						
						var po:Number = (event.delta_time/tpqn) * spqn;
						mustPlayNote = secs >= po;
						if (!mustPlayNote) {								
							break; //do not carry on reading events, it is not necesary, they can be only played in the future.				
						}						
						if (event.category == MidiMessage.CATEGORY_MIDI_CONTROL && (event.type == 9 || event.type == 8 || event.type == 0xC)) { //do we need to play this note event? 9 meand press note and 8 release note.
							if (event.type == 0xC) { //set sample index
								instruments[event.channel].setSampleIndex(event.param1);								
							} else if (event.type == 9) {									
								//note = getNodeByMidiNote(event.channel, event.param1);						
								note = instruments[event.channel].getKey(transpose[event.channel] + 3);																
								note.pressKey(event.param2); //hold down key.
								if (!isPlayingNote(note)) { //do not add the note twice.									
									playingNotes.push(note);								
									totalPlayingNotes++;								
								}
							}
						}
						trk.removeEvent(0);
						t--; //refresh total events.
						secs -= po;
						acc_time[n] = int(secs / STS);						
					}
				}
				
				sample.x = sample.y = 0; //silence sample								
				n = 0;
				while (n < totalPlayingNotes) {
					note_sample = playingNotes[n].getSample();
					if (note_sample) {
						vol = Math.max(note_sample.x, note_sample.y);
						if (vol > maxVel[playingNotes[n].channel]) {
							maxVel[playingNotes[n].channel] = vol;
						}
						vol = vols[playingNotes[n].channel].value * vols[6].value;
						sample.x += note_sample.x * vol;
						sample.y += note_sample.y * vol;
						n++;
					} else {
						playingNotes.splice(n, 1);
						totalPlayingNotes--;
					}
				}
				e.data.writeFloat(sample.x); 
				e.data.writeFloat(sample.y);
			}
			
			//update volume levels on mixer.
			maxVelAll = 0; //MASTER, maximum velocity (volume) of all channels.
			for (i = 0; i < 6; i++) {
				if (vols[i].volume < maxVel[i]){
					vols[i].volume = maxVel[i];
					vols[i].animateVolume();
					if (maxVelAll < maxVel[i]){
						maxVelAll = maxVel[i];
					}
				}
			}
			if (vols[6].volume < maxVelAll){
				vols[6].volume = maxVelAll; //master
				vols[6].animateVolume();
			}
		}
			
		//------------------- DYNAMIC AUDIO GENERATION [END]--------------------//
		
		private function stopSound(e:Event):void {
			removeEventListener(Event.ENTER_FRAME, stopSound);
			click(null);			
		}
	}
	
}
