package com.yoambulante.midifiles {
	/**
	 * MidiTrack contains all events related to the track, note that events channels are independant of the track
	 * @author Alex Nino (yoambulante.com)
	 */
	public class MidiTrack {
		private var _events:Vector.<MidiMessage>;
		private var _totalEvents:uint;	
		
		/**
		 * creates a new midi track which contents all different kind of midi events.
		 * @param	copyFrom copy all events from already created instance of MidiTrack (optional)
		 */
		public function MidiTrack(copyFrom:MidiTrack = null) {			
			_totalEvents = 0;
			_events = new Vector.<MidiMessage>();
			if (copyFrom as MidiTrack) {
				for (var i:int = 0; i < copyFrom.totalEvents; i++) {
					this.addEvent(copyFrom.getEvent(i).clone());
				}
			}
		}
		
		/**
		 * used by the main parser for adding event to the list.
		 * @param	e
		 */
		public function addEvent(e:MidiMessage):void {
			_events.push(e);			
			_totalEvents++;
		}
		
		/**
		 * removes a event from list
		 * @param	index must be a uint &gt;= 0 and &lt; totalEvents
		 * @return return null if index is equal or higher than totalEvents
		 */
		public function removeEvent(index:uint):void {
			if (index < _totalEvents) {
				_events.splice(index, 1);
				_totalEvents--;
			}
		}
		
		/**
		 * gets a event from list
		 * @param	index must be a uint &gt;= 0 and &lt; totalEvents
		 * @return return null if index is equal or higher than totalEvents
		 */
		public function getEvent(index:uint):MidiMessage {
			if (index < _totalEvents) {
				return _events[index];
			}
			return null;
		}
		
		/**
		 * duplicate current track, note it also duplicates its event messages inside track.
		 */
		public function clone():MidiTrack {
			var r:MidiTrack = new MidiTrack();
			for (var i:int = 0; i < totalEvents; i++) {
				r.addEvent(getEvent(i).clone());
			}
			return r;
		}
		
		/**
		 * returns total amount of available events in list
		 */
		public function get totalEvents():uint { return _totalEvents; }	
	}

}